-- ********************************************************************************
-- Copyright  Paul Pinault, Disk91 2005-06.
--
-- This program is free software/hardware; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- See : http://www.gnu.org/licenses/gpl.txt
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
--
-- Paul Pinault, Disk91 nor Simple-CPU terms can't be used in the way to promote
-- nor legitimate a derived product without the Author authorization.
--
-- =================================================================================
-- Using This software made your own application to become a free-software too
-- considered under GPL. In this case you will have to publish your modificated
-- VHDL code at least on a public website.
-- Perhaps, it's possible to contact the author at ( dpl AT disk91.com ) to obtain
-- the way to have a license spcific distribution of it with or without any fee
-- depending on your project.
-- =================================================================================
--
-- Ce programme est un logiciel / matriel libre, vous pouvez le redistribuer et/ ou
-- le modifier selon les termes de la GNU  General Public License comme publi par
-- la Free Software Fundation ; tablie en version 2 ou suprieure.
--
-- La license en Franais peut tre consultable sur le site suivant :
-- http://fsffrance.org/gpl/gpl-fr.fr.html
--
-- CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE
-- QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION
-- COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT
-- EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE
-- DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE
-- TOUT OU PARTIE DE CE PROGRAMME.
--
-- Les noms Paul Pinault, Disk91 ou Simple-CPU ne saurait tre utiliss dans le but
-- de promouvoir ou de lgitimer un produit driv de ce programme sans accord pralable
-- de l'auteur.
-- ********************************************************************************
--------------------------------------------------------------------------------
-- Company:        www.simple-cpu.com
-- Engineer:       Paul Pinault
--
-- Create Date:    19:47:00 09/23/05
-- Design Name:    sequenceur.vhdl - SC91 alu
--                            http://www.simple-cpu.com/instruction-chargement-valeur-directe.php
-- Module Name:    sc91-A - Behavioral
-- Project Name:   sc91-A
-- Target Device:  independant
-- Tool versions:
-- Description:    simple-cpu arithmetical and logical unit
--
-- Dependencies:   none
--
-- Revision:       1.0
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library core_lib;
use core_lib.utils.ALL;

library work;
use work.instruction_set.ALL;


-- *****************************************************************************
-- SEQUENCEUR ENTITY DESCRIPTION
-- *****************************************************************************
entity sc91_sequenceur is
   port (

        RST    : IN   std_logic;                         -- reset signal (Active High AH)
        CLK    : IN   std_logic;                         -- system clock

        -- Instructions
        IVV    : OUT  std_logic_vector(31 downto 0);     -- instruction bit for ALU
        INSTR  : IN   std_logic_vector(31 downto 0);     -- Intruction read from memory
        IN_RDY : IN   std_logic;                         -- Instruction is ready - insert a bubble if no (AH)

        -- Specific Data memory insterface
        DA_RDY : IN   std_logic;                         -- Data is ready - stay on the same instruction if no (AH)
                                                         -- ALU and MEMORY are READY
        -- Specific control signals
        ALU_SELECT: OUT  t_recoding;                     -- Indicate to ALU what output to choose (instruction recoding)
        ALU_ROUTE : OUT  t_route;                        -- choose ALU output depending on instruction

        -- Register selection
        RNN    : OUT  std_logic_vector(4 downto 0);      -- first source operand
        RMM    : OUT  std_logic_vector(4 downto 0);      -- second source operand
        RDD    : OUT  std_logic_vector(4 downto 0);      -- destination operand
        LDD    : OUT  std_logic;                         -- load RDD register with DDD value (AH)

        -- Control PC
        PCDIR  : OUT  std_logic_vector(31 downto 0);    -- Program counter register direct input
        PC_R00 : OUT  std_logic_vector( 1 downto 0);    -- bit 1 control PC output as 1 => PCDIR 0 => R00
                                                        -- bit 0 control PC input  as 1 => Ddd   0 => PC + 4
        PC_LOAD: OUT  std_logic;                        -- load PC with its input value (AH)

        -- Specific flag signals
        FL     : IN   std_logic_vector(15 downto 0);     -- computed flags for ALU and Instruction
        LFL    : OUT  std_logic;                         -- update Flags with current result (AH)

        -- Specific Stack pointer signals
        SPP    : OUT  std_logic;                         -- store in SP (R01) the current value + 4   (AH)
        SPM    : OUT  std_logic;                         -- put on R01 output current value - 4       (AH)

        -- not use
        NU        : OUT std_logic                       -- not used, just that last line doesn't need ";"

   );
end sc91_sequenceur;

architecture arch_sc91_sequenceur of sc91_sequenceur is

   -- ==========================================================================================
   -- SOME CONSTANTS
   constant INSTR_RDD_FIELD_TOP : INTEGER := 23;        -- Higher bit for RDD value in instruction encoding
   constant INSTR_RNN_FIELD_TOP : INTEGER := 18;        -- Higher bit for RNN value in instruction encoding
   constant INSTR_RMM_FIELD_TOP : INTEGER := 13;        -- Higher bit for RMM value in instruction encoding
   constant INSTRUCTION_NOP : std_logic_vector(31 downto 0) := X"FF000000";     -- NOP Instruction encoding


   -- ==========================================================================================
   -- MISC SIGNALS
   signal instruction : std_logic_vector(31 downto 0);   -- Current instruction value
   signal i_INSTR     : std_logic_vector(31 downto 0);   -- Current instruction value
   signal i_ADDR_LDD  : std_logic;
   signal i_LOAD      : std_logic;
   signal i_destIsR00 : std_logic;                       -- Destination Register is R00
   signal i_OPCODE    : std_logic_vector(7 downto 0);
   signal i_PCR00     : std_logic_vector(1 downto 0);
   signal i_ALUSELECT : t_recoding;

begin
   -- ==========================================================================================
   -- CURRENT INSTRUCTION REGISTER
   PC_LOAD <= '0' when IN_RDY  = '0'  else               -- don't change PC while memory is not ready or ALU not ready
              '0' when DA_RDY  = '0'  else
              '0' when i_PCR00 = "01" else               -- or when R00 just changed
              '1';

   i_INSTR <= INSTRUCTION_NOP when ( IN_RDY     = '0' ) else
              INSTRUCTION_NOP when ( i_PCR00(1) = '1' ) else -- after a write into R00 from DDD, we have to wait an
              INSTR;                                         -- extra cycle to wait code memory to be ready to serve

   i_LOAD  <= DA_RDY;                                        -- ALU and MEMORY ARE READY
   i_rinstr : cl_reg generic map ( WIDTH => 32 )
                     port map ( D => i_INSTR, Q => instruction, CLK => CLK, LOAD => i_LOAD, RESET => RST);

   IVV <= instruction;


   -- For ADDR Load Drr only when mode is not CMP
   i_ADDR_LDD <= '0' when instruction(IVV_ADDR_MODE_TOP downto IVV_ADDR_MODE_TOP-2) = ADDR_MODE_CMP else
                 '1';
   -- ==========================================================================================
   -- INSTRUCTION DECODING
   i_OPCODE <= instruction(31 downto 24);
   process ( instruction, i_OPCODE, DA_RDY, i_ADDR_LDD, i_destIsR00 )
   begin
      -- Default values
      SPP         <= '0';
      SPM         <= '0';
      LFL         <= '1';
      i_PCR00(1)  <= i_destIsR00;                           -- load from DDD when destination is R00
      i_PCR00(0)  <= '0';                                   -- output R00 as instruction address.
      LDD         <= '0';
      i_ALUSELECT <= (others => '0');

      -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      -- Instuction JMP
      if     ( i_OPCODE(7 downto 6) = "00" ) then
         i_PCR00(0)                 <= '1';                         -- Route direct address to instruction address
         LFL                        <= '0';
         i_ALUSELECT(SELECT_JMP )   <= '1';
         ALU_ROUTE                  <= ROUTE_DEFAULT;

      -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      -- Instruction CALL
      elsif  ( i_OPCODE(7 downto 6) = "01" ) then
         i_ALUSELECT(SELECT_CALL)   <= '1';
         LFL                        <= '0';
         LDD                        <= DA_RDY;         -- Save SPP only when DA_RDY
         SPP                        <= '1';
         i_PCR00(1)                 <= '0';            -- Be sure to have 0
         i_PCR00(0)                 <= '1';            -- Route direct address to instruction address
         ALU_ROUTE                  <= ROUTE_DEFAULT;

      -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      -- Instuction JB
      elsif  ( i_OPCODE(7 downto 5) = "110" ) then
         i_PCR00(1)                 <= '1';   -- load from DDD
         LFL                        <= '0';
         i_ALUSELECT(SELECT_JB)     <= '1';
         ALU_ROUTE                  <= ROUTE_JB;

      -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      -- Instruction ADDd
      elsif  ( i_OPCODE(7 downto 5) = "101" ) then
         i_ALUSELECT(SELECT_ADDd)   <= '1';
         LDD                        <= '1';
         ALU_ROUTE                  <= ROUTE_ADD;

      else
         case i_OPCODE is
            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction LOADd
            when OPCODE_LOADD  =>
               LDD                       <= '1';             -- load result into RDD
               i_ALUSELECT(SELECT_LOADD) <= '1';
               ALU_ROUTE                 <= ROUTE_LOADd;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction LOADr
            when OPCODE_LOADR  =>
               LDD                       <= '1';             -- load result into RDD
               i_ALUSELECT(SELECT_LOADR) <= '1';
               ALU_ROUTE                 <= ROUTE_LOADx;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction LOADm
            when OPCODE_LOADM  =>
               LDD                       <= DA_RDY;    -- Save value only when DA_RDY
               i_ALUSELECT(SELECT_LOADM) <= '1';
               ALU_ROUTE                 <= ROUTE_LOADx;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction POP
            when OPCODE_POP  =>
               LDD                       <= DA_RDY;    -- Save value only when DA_RDY
               i_ALUSELECT(SELECT_POP)   <= '1';
               SPM                       <= '1';       -- Decrement R02
               ALU_ROUTE                 <= ROUTE_LOADx;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction NOP
            when OPCODE_NOP =>
               LFL                        <= '0';
               i_PCR00(1)                 <= '0';      -- Be sure to have 0
               i_ALUSELECT(SELECT_NOP)    <= '1';
               ALU_ROUTE                  <= ROUTE_DEFAULT;


            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction AND
            when OPCODE_AND =>
               LDD                        <= '1';
               i_ALUSELECT(SELECT_AND)    <= '1';
               ALU_ROUTE                  <= ROUTE_AND;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction OR
            when OPCODE_OR =>
               LDD                        <= '1';
               i_ALUSELECT(SELECT_OR)     <= '1';
               ALU_ROUTE                  <= ROUTE_OR;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction XOR
            when OPCODE_XOR =>
               LDD                        <= '1';
               i_ALUSELECT(SELECT_XOR)    <= '1';
               ALU_ROUTE                  <= ROUTE_XOR;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction JMPR
            when OPCODE_JMPR =>
               i_PCR00(1)                 <= '1';         -- load from DDD
               LFL                        <= '0';
               i_ALUSELECT(SELECT_JMPR)   <= '1';         -- route Rmm on DDD
               ALU_ROUTE                  <= ROUTE_DEFAULT;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction CALLR
            when OPCODE_CALLR =>
               i_PCR00(1)                 <= '1';         -- load from DDD
               LFL                        <= '0';
               LDD                        <= DA_RDY;      -- Save SPP only when DA_RDY
               SPP                        <= '1';
               i_ALUSELECT(SELECT_CALLR)  <= '1';         -- route Rmm on DDD
               ALU_ROUTE                  <= ROUTE_DEFAULT;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction CPYB
            when OPCODE_CPYB =>
               LDD                        <= '1';
               i_ALUSELECT(SELECT_CPYB)   <= '1';
               i_PCR00(1)                 <= '0';           -- Be sure to have 0
               ALU_ROUTE                  <= ROUTE_CPYB;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction ADDR / CMP
            when OPCODE_ADDR =>
               i_ALUSELECT(SELECT_ADDR)   <= '1';
               LDD                        <= i_ADDR_LDD;
               ALU_ROUTE                  <= ROUTE_ADD;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction MUL
            when OPCODE_MUL =>
               i_ALUSELECT(SELECT_MUL)    <= '1';
               LDD                        <= '1';
               ALU_ROUTE                  <= ROUTE_MUL;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction DIV
            when OPCODE_DIV =>
               i_ALUSELECT(SELECT_DIV)    <= '1';
               LDD                        <= DA_RDY;
               ALU_ROUTE                  <= ROUTE_DIV;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction ROT
            when OPCODE_ROT =>
               i_ALUSELECT(SELECT_ROT)    <= '1';
               LDD                        <= '1';
               ALU_ROUTE                  <= ROUTE_ROT;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction PUSH
            when OPCODE_PUSH =>
               i_ALUSELECT(SELECT_PUSH)   <= '1';
               LFL                        <= '0';
               LDD                        <= DA_RDY;         -- Save SPP only when DA_RDY
               SPP                        <= '1';
               i_PCR00(1)                 <= '0';            -- Be sure to have 0
               ALU_ROUTE                  <= ROUTE_ADD;

            -- ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            -- Instruction STORE
            when OPCODE_STORE =>
               i_ALUSELECT(SELECT_STORE)  <= '1';
               LFL                        <= '0';
               LDD                        <= DA_RDY;
               i_PCR00(1)                 <= '0';           -- Be sure to have 0 STORE R00 forbiden
               ALU_ROUTE                  <= ROUTE_DEFAULT;


            when others      =>
               i_PCR00(1)  <= '0';           -- Be sure to have 0
               ALU_ROUTE                  <= ROUTE_DEFAULT;

         end case;
      end if;
   end process;
   ALU_SELECT <= i_ALUSELECT;

   -- ==========================================================================================
   -- REGISTER CONTROL LOGIC
   i_destIsR00 <= '1' when instruction(INSTR_RDD_FIELD_TOP downto INSTR_RDD_FIELD_TOP-4) = "00000" else '0';
   RDD <= instruction(INSTR_RDD_FIELD_TOP downto INSTR_RDD_FIELD_TOP-4);

   RNN <= "00000" when i_ALUSELECT(SELECT_CALL) = '1' else                                         -- PC on DNN
          instruction(INSTR_RNN_FIELD_TOP downto INSTR_RNN_FIELD_TOP-4);

   RMM <= instruction(INSTR_RDD_FIELD_TOP downto INSTR_RDD_FIELD_TOP-4) when i_ALUSELECT(SELECT_LOADD) = '1' else
          instruction(INSTR_RDD_FIELD_TOP downto INSTR_RDD_FIELD_TOP-4) when i_ALUSELECT(SELECT_CPYB)  = '1' else
          "00000"                                                       when i_ALUSELECT(SELECT_JB   ) = '1' else
          instruction(INSTR_RMM_FIELD_TOP downto INSTR_RMM_FIELD_TOP-4);

   PCDIR(31 downto  2) <= instruction(29 downto 0);
   PCDIR( 1 downto  0) <= "00";
   PC_R00 <= i_PCR00;

end arch_sc91_sequenceur;
